@@ -1,5 +1,5 @@
-__version__ = "0.0.32"
-__keywords__ = ["python cgi apache"]
+__version__ = "0.0.33"
+__keywords__ = ["python cgi apache stream file cache server proxy http"]
 
 
 # if not __version__.endswith(".0"):
