@@ -1,4 +1,5 @@
 import os
+import re
 import sys
 import json
 import math
@@ -9,7 +10,7 @@ from dateutil.tz import tzoffset, tzlocal
 import http.client
 from http.cookies import SimpleCookie
 from cgi import FieldStorage
-from omnitools import dt2yyyymmddhhmmss, ApacheHeadersDict, str2html
+from omnitools import dt2yyyymmddhhmmss, HeadersDict, ApacheHeadersDict, str2html, encodeURIComponent
 
 
 _SERVER = dict(os.environ)
@@ -18,6 +19,7 @@ for k in list(_SERVER.keys()):
         _SERVER[k.replace("REDIRECT_", "")] = _SERVER[k]
 arguments = FieldStorage(environ=_SERVER)
 arguments = {k: [_.value for _ in arguments[k]] if isinstance(arguments[k], list) else arguments[k].value for k in arguments}
+_SERVER = HeadersDict(_SERVER)
 _GET = arguments
 _POST = arguments
 _SESSION = SimpleCookie()
@@ -63,20 +65,24 @@ PRINTED = {
 }
 
 
-def obj_to_bytes(obj):
+def obj_to_bytes(obj, html: bool = True):
     if isinstance(obj, str):
         obj = obj.encode()
     elif not isinstance(obj, bytes):
         try:
-            obj = str2html(json.dumps(obj, indent=2))
+            obj = json.dumps(obj, indent=2)
+            if html:
+                obj = str2html(obj)
         except:
-            obj = str2html(str(obj))
+            obj = str(obj)
+            if html:
+                obj = str2html(obj)
         obj = obj.encode()
     return obj
 
 
 def log(obj, fp: str = None):
-    obj = obj_to_bytes(obj)
+    obj = obj_to_bytes(obj, html=False)
     now = dt2yyyymmddhhmmss().encode()
     open(fp or "log.log", "ab").write(now+b" "+obj+b"\n")
 
@@ -123,9 +129,8 @@ def print(obj = "", end=b"\n"):
         __response["cache"] += end
 
 
-def traceback(tag_name: str = "code", class_name: str = "traceback", style: str = "", limit=None, chain=True):
-    return "<{tag} class='{}' style='{}'>{}</{tag}>".format(
-        class_name,
+def traceback(tag_name: str = "code", style: str = "", limit=None, chain=True):
+    return "<{tag} style='{}'>{}</{tag}>".format(
         style,
         str2html(_traceback.format_exc(limit, chain)),
         tag=tag_name
@@ -164,7 +169,7 @@ def _should_return_304(fp: str = None):
         return False
     lastmodified = math.floor(os.path.getmtime(fp))
     lastmodified2 = datetime.datetime.fromtimestamp(lastmodified, tz=tzoffset(None, 0))
-    set_header("Last-Modified", lastmodified2.strftime('%a, %d %b %Y %H:%M:%S GMT'))
+    set_header("Last-Modified", lastmodified2.strftime("%a, %d %b %Y %H:%M:%S GMT"))
     if "If-Modified-Since" in _HEADERS:
         ims = email.utils.parsedate(_HEADERS["If-Modified-Since"])[:6]
         if ims:
@@ -191,9 +196,9 @@ def _cache_control():
 
 
 def _should_read_from_cache_file():
-    last_modiied = 999
+    last_modified = 999
     if_modified_since = 123
-    if if_modified_since >= last_modiied:
+    if if_modified_since >= last_modified:
         # cache = pickle.loads(open(_SERVER["SCRIPT_NAME"]+"cache", "rb").read())
         # set_status(cache["status_code"])
         # for k, v in cache["headers"]:
@@ -210,6 +215,66 @@ def _write_to_cache_file(cache):
     return
 
 
+def generate_range_headers(fp, disposition: str = "inline"):
+    if not os.path.isfile(fp):
+        set_status(404)
+        raise FileNotFoundError
+    range = [0, None]
+    if "Range" in _HEADERS:
+        m = re.match(r"bytes=(\d+)-(\d+)?$", _HEADERS["Range"])
+        if m:
+            start, end = [int(x) if x else x for x in m.groups()]
+            if end and end < start:
+                set_status(416)
+                raise ValueError("end [{}] < start [{}]".format(end, start))
+            range = [start, end]
+    try:
+        fo = open(fp, "rb")
+    except:
+        set_status(500)
+        raise IOError("cannot open file")
+    filename = os.path.basename(fp)
+    size = os.path.getsize(fp)
+    if range[0] >= size:
+        set_status(416)
+        raise ValueError("start [{}] > file size [{}]".format(range[0], size))
+    if not range[1] or range[1] >= size:
+        range[1] = size-1
+    if range[0] != 0 or range[1] < size-1:
+        set_status(206)
+    length = range[1]-range[0]+1
+    try:
+        filename.encode("ascii")
+        filename = 'filename="{}"'.format(filename)
+    except UnicodeEncodeError:
+        filename = "filename*=utf-8''{}".format(encodeURIComponent(filename))
+    set_header("Accept-Range", "bytes")
+    set_header("Content-Type", "application/octet-stream")
+    set_header("Content-Disposition", "{}; {}".format(disposition, filename))
+    set_header("Content-Range", "bytes {}-{}/{}".format(range[0], range[1], size))
+    set_header("Content-Length", str(length))
+    return fo, range
+
+
+def print_file(fo, buf_size: int = 32 * 1024, range = None):
+    if not hasattr(fo, "tell") or not hasattr(fo, "read") or not hasattr(fo, "seek") or not hasattr(fo, "close"):
+        raise IOError("{} is not file object".format(fo))
+    if not range:
+        fo.seek(0, 2)
+        range = [0, fo.tell()]
+    fo.seek(range[0])
+    while True:
+        buf = min(buf_size, range[1] + 1 - fo.tell())
+        buffer = fo.read(buf)
+        buf = None
+        if not buffer:
+            break
+        print(buffer, end="")
+        buffer = None
+        flush()
+    fo.close()
+
+
 def execute(
         method: str = "get", cacheable: bool = False,
         cache_ctrl = _cache_control,
@@ -247,7 +312,7 @@ def execute(
                     tb = traceback(**(traceback_kwargs or {}))
                 else:
                     tb = "<h1>500 Internal Server Error</h1><br/><p>HTML stack trace is disabled.<br/>Check traceback log.</p>"
-                __response["content"] = tb.encode()
+                __response["content"] = obj_to_bytes(tb)
                 try:
                     set_status(500)
                 except:
